@@ -4,18 +4,16 @@ from copy import copy
 from dataclasses import dataclass, field, replace
 from functools import wraps
 from itertools import chain
+from pprint import pformat
 from typing import Union, Type, TypeVar, Callable, Dict, Any
 
 import cloudpickle
 import pinject
-from cytoolz import merge, valmap
-from expression import Some, Nothing
-from expression.collections import Map
-from loguru import logger
+from cytoolz import merge, valmap, itemmap
 from makefun import create_function, wraps
 from pampy import match
 from pinject import BindingSpec
-from pinject.scoping import SingletonScope
+from pinject.scoping import SingletonScope, BindableScopes, SINGLETON
 from pinject_design.di.design import Bind, FunctionProvider, ProviderTrait, InjectedProvider, PinjectConfigure, \
     PinjectProvider, ensure_self_arg, PinjectBind
 from pinject_design.di.graph import ExtendedObjectGraph
@@ -23,21 +21,47 @@ from pinject_design.di.injected import Injected
 from returns.result import safe, Failure
 from tabulate import tabulate
 
+from pinject_design.di.session import SessionScope
+
 prototype = pinject.provides(in_scope=pinject.PROTOTYPE)
 # is it possible to create a binding class which also has an ability to dynamically add..?
 # yes. actually.
 T = TypeVar("T")
 
 
+def rec_valmap(f, tgt: dict):
+    res = dict()
+    for k, v in tgt.items():
+        if isinstance(v, dict):
+            res[k] = rec_valmap(tgt[k], v)
+        else:
+            res[k] = f(v)
+    return res
+
+
+def rec_val_filter(f, tgt: dict):
+    res = dict()
+    for k, v in tgt.items():
+        if isinstance(v, dict):
+            res[k] = rec_val_filter(f, v)
+        elif f(v):
+            res[k] = v
+    return res
+
+
 def check_picklable(tgt: dict):
     cloud_dumps_try = safe(cloudpickle.dumps)
     cloud_loads_try = safe(cloudpickle.loads)
     res = cloud_dumps_try(tgt).bind(cloud_loads_try)
 
     if isinstance(res, Failure):
-        target_check = Map.of(**valmap(cloud_dumps_try, tgt))
-        logger.error(
-            "failed pickling:\n" + tabulate(target_check.filter(lambda k, v: isinstance(v, Failure)).to_list()))
+        # target_check = valmap(cloud_dumps_try, tgt)
+        rec_check = rec_valmap(lambda v: (cloud_dumps_try(v), v), tgt)
+        failures = rec_val_filter(lambda v: isinstance(v[0], Failure), rec_check)
+        # failures = [(k, v, tgt[k]) for k, v in target_check.items() if isinstance(v, Failure)]
+
+        from loguru import logger
+        logger.error(f"Failed to pickle target: {pformat(failures)}")
         logger.error(f"if the error message contains EncodedFile pickling error, "
                      f"check whether the logging module is included in the target object or not.")
         raise RuntimeError("this object is not picklable. check the error messages above.")
@@ -111,10 +135,7 @@ def get_class_aware_args(f):
 
 
 def getitem_opt(o, k):
-    if k in o:
-        return Some(o[k])
-    else:
-        return Nothing
+    return safe(o.__getitem__)(k)
 
 
 def to_readable_name(o):
@@ -148,8 +169,8 @@ def get_dict_diff(a: dict, b: dict):
     Subject = try_import_subject()
     for k in all_keys:
 
-        ak = getitem_opt(a, k).map(to_readable_name).or_else(Some(None)).value
-        bk = getitem_opt(b, k).map(to_readable_name).or_else(Some(None)).value
+        ak = getitem_opt(a, k).map(to_readable_name).value_or(None)
+        bk = getitem_opt(b, k).map(to_readable_name).value_or(None)
         flag = match((ak, bk),
                      (Subject, Subject), lambda a, b: True,
                      # (np.ndarray, np.ndarray), lambda a, b: (a != b).any(),
@@ -214,11 +235,12 @@ class Design:
                      )
 
     def _merge_multi_binds(self, src, dst):
-        src = Map.of(**src)
-        dst = Map.of(**dst)
+        # src = Map.of(**src)
+        # dst = Map.of(**dst)
         keys = src.keys() | dst.keys()
         multi = {k: (
-                src.try_find(k).default_value([]) + dst.try_find(k).default_value([])
+                getitem_opt(src, k).value_or([]) +
+                getitem_opt(dst, k).value_or([])
         ) for k in keys}
         return multi
 
@@ -260,11 +282,13 @@ class Design:
         x = self
         for k, v in kwargs.items():
             if isinstance(v, type):
+                from loguru import logger
                 logger.warning(f"{k} is bound to class {v} with 'bind_instance' do you mean 'bind_class'?")
             x = x.bind(k).to_instance(v)
         return x
 
     def bind_provider(self, **kwargs: Union[Callable, Injected]):
+        from loguru import logger
         x = self
         for k, v in kwargs.items():
             # logger.info(f"binding provider:{k}=>{v}")
@@ -282,6 +306,7 @@ class Design:
         return x
 
     def bind_class(self, **kwargs):
+        from loguru import logger
         x = self
         for k, v in kwargs.items():
             if isinstance(v, Injected):
@@ -295,13 +320,15 @@ class Design:
         modules = self.modules + (modules or [])
         classes = self.classes + (classes or [])
         # logger.info(f"to_graph:\n\t{pformat(modules)}\n\t{pformat(classes)}")
-        return ExtendedObjectGraph(
-            self,
-            pinject.new_object_graph(
-                modules=modules,
-                binding_specs=[self.to_binding_spec()],
-                classes=classes
-            ))
+        g = pinject.new_object_graph(
+            modules=modules,
+            binding_specs=[self.to_binding_spec()],
+            classes=classes
+        )
+        g._obj_provider._bindable_scopes = BindableScopes(
+            id_to_scope={SINGLETON: SessionScope()}
+        )
+        return ExtendedObjectGraph(self, g)
 
     def run(self, f, modules=None, classes=None):
         return self.to_graph(modules, classes).run(f)
@@ -441,11 +468,17 @@ class Design:
         """Generate a BindingSpec class for pinject and returns its instance."""
 
         design = self.build()
-        bindings = Map.of(**design.bindings).map(lambda k, v: v.to_pinject_binding())
-        configures = bindings.filter(lambda k, v: isinstance(v, PinjectConfigure))
-        providers = bindings.filter(lambda k, v: isinstance(v, PinjectProvider)).map(
-            lambda k, v: self._ensure_provider_name(k, v.method)
-        )
+        bindings = {k: v.to_pinject_binding() for k, v in design.bindings.items()}
+        configures = {k: v for k, v in bindings.items() if isinstance(v, PinjectConfigure)}
+        # configures = bindings.filter(lambda k, v: isinstance(v, PinjectConfigure))
+        providers = {k: v for k, v in bindings.items() if isinstance(v, PinjectProvider)}
+        providers = {k: self._ensure_provider_name(k, v.method) for k, v in providers.items()}
+
+        # providers = itemmap(lambda t: self._ensure_provider_name(t[0], t[1].method), providers)
+
+        # providers = bindings.filter(lambda k, v: isinstance(v, PinjectProvider)).map(
+        #     lambda k, v: self._ensure_provider_name(k, v.method)
+        # )
 
         def configure(self, bind):
             for c, v in dict(configures).items():
@@ -462,6 +495,7 @@ class Design:
 
     def _ensure_provider_name(self, k, method):
         """set appropriate name for provider function to be recognized by pinject"""
+        from loguru import logger
         name = f"provide_{k}"
         if not method.__name__ == name:
             # there are cases where you cannot directly set __name__ attribute.
@@ -526,7 +560,9 @@ class Design:
 
         def create_impl(tgt_providers, tgt_dependencies):
             def f_impl(**kwargs):
-                logger.info(f"{f_signature} called with {list(kwargs.keys())}")
+                # from loguru import logger
+
+                # logger.info(f"{f_signature} called with {list(kwargs.keys())}")
                 values = []
                 for provider, ds in zip(tgt_providers, tgt_dependencies):
                     p_deps = {k: kwargs[k] for k in ds}
@@ -546,6 +582,26 @@ class Design:
         d = get_dict_diff(self.bindings, other.bindings)
         return d
 
+    def inspect_picklability(self):
+        from loguru import logger
+        logger.info(f"checking picklability of bindings")
+        check_picklable(self.bindings)
+        logger.info(f"checking picklability of multi-binds")
+        check_picklable(self.multi_binds)
+        logger.info(f"checking picklability of modules")
+        check_picklable(self.modules)
+        logger.info(f"checking picklability of classes")
+        check_picklable(self.classes)
+
+    def add_modules(self, *modules):
+        return self + Design(modules=list(modules))
+
+    def add_classes(self, *classes):
+        return self + Design(classes=list(classes))
+
+
+EmptyDesign = Design()
+
 
 @dataclass
 class DirectPinjectProvider(Bind):
@@ -591,6 +647,7 @@ def _patched_provide(self, binding_key, default_provider_fn):
 
 
 def patch_pinject_singleton_scope():
+    from loguru import logger
     logger.warning(f"patching pinject's SingletonScope to produce better exception")
     SingletonScope.provide = _patched_provide
     logger.warning(f"patching done")
@@ -616,6 +673,7 @@ def _get_external_type_name(thing):
 
 def patch_pinject_get_external_type_name():
     import pinject.locations
+    from loguru import logger
     logger.warning(f"patching pinject's _get_external_type_name to accept pickled function")
     pinject.locations._get_external_type_name = _get_external_type_name
 
